@@ -14,12 +14,17 @@ int APIENTRY WinMain(HINSTANCE hInstance,
   char basePath[MAX_PATH];
   char classPath[MAX_PATH*2];
   char libPath[MAX_PATH];
+  char tempPath[MAX_PATH];
+  char tmpPath[MAX_PATH];
   char *args[5];
 
   iRet = GetModuleFileName(NULL, path, sizeof(path));
 
   _splitpath_s( path, szDrive, _MAX_DRIVE, szDir, _MAX_DIR, szFileName, _MAX_FNAME, szExt, _MAX_EXT );
 
+  char tempF[MAX_PATH];
+  GetEnvironmentVariable("TEMP",tempF,MAX_PATH);
+
   strcpy_s(basePath, szDrive);
   strcat_s(basePath, szDir);
   strcpy_s(classPath, "CLASSPATH=");
@@ -30,8 +35,12 @@ int APIENTRY WinMain(HINSTANCE hInstance,
   strcpy_s(libPath, "PATH=");
   strcat_s(libPath, basePath);
   strcat_s(libPath, "acerola3d");
+  strcpy_s(tempPath, "TEMP=");
+  strcat_s(tempPath, tempF);
+  strcpy_s(tmpPath, "TMP=");
+  strcat_s(tmpPath, tempF);
 
-  char *my_env[] = { classPath, libPath, NULL };
+  char *my_env[] = { classPath, libPath, tempPath, tmpPath, NULL };
 
   if ((!lpCmdLine)||(!strlen(lpCmdLine))) {
     args[0] = "javaw.exe";
