@@ -13,8 +13,6 @@ class A3eActionEditor {
     Catalog catalog;
     A3eActionEditorGUI gui;
     A currentAction;
-    Skeleton currentSkeleton;
-    HashMap<String,CenteredVRML> vrmls = new HashMap<String,CenteredVRML>();
 
     A3eActionEditor(A3Editor e) {
         editor = e;
@@ -23,78 +21,14 @@ class A3eActionEditor {
         setCurrent(catalog.a3.getA().get(0));
         gui = new A3eActionEditorGUI(this,editor);
     }
+    String getBVHPath() {
+        return currentAction.getBvh();
+    }
+    String getBVHURL() {
+        return fileManager.path2UrlString(currentAction.getBvh());
+    }
     void setCurrent(A a) {
         currentAction = a;
-        String bvhPath = currentAction.getBvh();
-        String bvhUrl = fileManager.path2UrlString(bvhPath);
-        if (bvhUrl!=null && !bvhPath.equals("none")) {
-            try {
-                currentSkeleton = new Skeleton(bvhUrl);
-            } catch (Exception e1) {
-                e1.printStackTrace();
-            }
-        } else {
-            currentSkeleton = null;
-        }
-        String s = currentAction.getOffset();
-        double x = Double.parseDouble(s.split("\\s")[0]);
-        double y = Double.parseDouble(s.split("\\s")[1]);
-        double z = Double.parseDouble(s.split("\\s")[2]);
-        Vector3d actionOffset = new Vector3d(x,y,z);
-        s = currentAction.getRot();
-        x = Double.parseDouble(s.split("\\s")[0])/180.0*Math.PI;
-        y = Double.parseDouble(s.split("\\s")[1])/180.0*Math.PI;
-        z = Double.parseDouble(s.split("\\s")[2])/180.0*Math.PI;
-        Quat4d actionQuat = Util.euler2quat(x,y,z);
-        double actionScale = currentAction.getScale();
-        if (currentSkeleton!=null) {
-            currentSkeleton.setLoc(actionOffset);
-            currentSkeleton.setQuat(actionQuat);
-            currentSkeleton.setScale(actionScale);
-        }
-        //VRML.clearCash();
-        vrmls.clear();
-        for (P p : currentAction.getP()) {
-            String path = p.getWrl();
-            String urlString = fileManager.path2UrlString(path);
-            try {
-                CenteredVRML vrml = new CenteredVRML(urlString);
-                Vector3d loc = new Vector3d();
-                Vector3d bvhOffset = null; //BVHデータから来る骨のオフセット
-                if (currentSkeleton!=null)
-                    bvhOffset = currentSkeleton.getOffset(p.getName());
-                else
-                    bvhOffset = new Vector3d();
-                s = p.getOffset();
-                x = Double.parseDouble(s.split("\\s")[0]);
-                y = Double.parseDouble(s.split("\\s")[1]);
-                z = Double.parseDouble(s.split("\\s")[2]);
-                Vector3d boneOffset = new Vector3d(x,y,z);//CATALOG.XMLから来る骨のオフセット
-                s = p.getRot();
-                x = Double.parseDouble(s.split("\\s")[0])/180.0*Math.PI;
-                y = Double.parseDouble(s.split("\\s")[1])/180.0*Math.PI;
-                z = Double.parseDouble(s.split("\\s")[2])/180.0*Math.PI;
-                Quat4d boneQuat = Util.euler2quat(x,y,z);//CATALOC.XMLから来る骨の回転
-                double boneScale = p.getScale();//CATALOG.XMLから来る骨のスケール
-
-                loc.add(boneOffset);
-                loc.add(bvhOffset);
-                loc.scale(actionScale);
-                loc = Util.trans(actionQuat,loc);
-                loc.add(actionOffset);
-                Quat4d quat = new Quat4d();
-                quat.mul(actionQuat,boneQuat);
-                double scale = boneScale*actionScale;
-
-                vrml.setQuat(quat);
-                vrml.setScale(scale);
-                vrml.setLoc2(loc);
-
-                vrmls.put(p.getName(),vrml);
-            } catch (Exception e) {
-                e.printStackTrace();
-            }
-        }
     }
     void changeCurrentAction(String an) {
         A ca = null;
@@ -236,12 +170,6 @@ class A3eActionEditor {
     String getCurrentActionName() {
         return currentAction.getAn();
     }
-    Skeleton getCurrentSkeleton() {
-        return currentSkeleton;
-    }
-    CenteredVRML[] getCurrentVRMLs() {
-        return vrmls.values().toArray(new CenteredVRML[0]);
-    }
     boolean checkActionNameDuplication(String an) {
         for (A a : catalog.a3.getA()) {
             if (a.getAn().equals(an))
@@ -292,91 +220,35 @@ class A3eActionEditor {
     void setDalsegno(double d) {
         currentAction.setDalsegno(d);
     }
-    double getRBX() {//右の吹き出しのX座標
-        String s = currentAction.getRightBalloonOffset();
-        return Double.parseDouble(s.split("\\s")[0]);
+    double[] getRightBalloonOffset() {//右の吹き出し
+        return string2vector2d(currentAction.getRightBalloonOffset());
     }
-    void setRBX(double x) {//右の吹き出しのX座標
-        String s = currentAction.getRightBalloonOffset();
-        currentAction.setRightBalloonOffset(""+x+" "+s.split("\\s")[1]);
+    void setRightBalloonOffset(double[] v) {//右の吹き出し
+        currentAction.setRightBalloonOffset(vector2d2string(v));
     }
-    double getRBY() {//右の吹き出しのY座標
-        String s = currentAction.getRightBalloonOffset();
-        return Double.parseDouble(s.split("\\s")[1]);
+    double[] getLeftBalloonOffset() {//左の吹き出し
+        return string2vector2d(currentAction.getLeftBalloonOffset());
     }
-    void setRBY(double y) {//右の吹き出しのY座標
-        String s = currentAction.getRightBalloonOffset();
-        currentAction.setRightBalloonOffset(s.split("\\s")[0]+" "+y);
+    void setLeftBalloonOffset(double[] v) {//左の吹き出し
+        currentAction.setLeftBalloonOffset(vector2d2string(v));
     }
-    double getLBX() {//左の吹き出しのX座標
-        String s = currentAction.getLeftBalloonOffset();
-        return Double.parseDouble(s.split("\\s")[0]);
+    double[] getTopBalloonOffset() {//上の吹き出し
+        return string2vector2d(currentAction.getTopBalloonOffset());
     }
-    void setLBX(double x) {//左の吹き出しのX座標
-        String s = currentAction.getLeftBalloonOffset();
-        currentAction.setLeftBalloonOffset(""+x+" "+s.split("\\s")[1]);
+    void setTopBalloonOffset(double v[]) {//上の吹き出し
+        currentAction.setTopBalloonOffset(vector2d2string(v));
     }
-    double getLBY() {//左の吹き出しのY座標
-        String s = currentAction.getLeftBalloonOffset();
-        return Double.parseDouble(s.split("\\s")[1]);
+    double[] getBottomBalloonOffset() {//下の吹き出し
+        return string2vector2d(currentAction.getBottomBalloonOffset());
     }
-    void setLBY(double y) {//左の吹き出しのY座標
-        String s = currentAction.getLeftBalloonOffset();
-        currentAction.setLeftBalloonOffset(s.split("\\s")[0]+" "+y);
+    void setBottomBalloonOffset(double v[]) {//上の吹き出し
+        currentAction.setBottomBalloonOffset(vector2d2string(v));
     }
-    double getTBX() {//上の吹き出しのX座標
-        String s = currentAction.getTopBalloonOffset();
-        return Double.parseDouble(s.split("\\s")[0]);
+    double[] getLabelOffset() {//ラベル
+        return string2vector2d(currentAction.getLabelOffset());
     }
-    void setTBX(double x) {//上の吹き出しのX座標
-        String s = currentAction.getTopBalloonOffset();
-        currentAction.setTopBalloonOffset(""+x+" "+s.split("\\s")[1]);
-    }
-    double getTBY() {//上の吹き出しのY座標
-        String s = currentAction.getTopBalloonOffset();
-        return Double.parseDouble(s.split("\\s")[1]);
-    }
-    void setTBY(double y) {//上の吹き出しのY座標
-        String s = currentAction.getTopBalloonOffset();
-        currentAction.setTopBalloonOffset(s.split("\\s")[0]+" "+y);
-    }
-    double getBBX() {//下の吹き出しのX座標
-        String s = currentAction.getBottomBalloonOffset();
-        return Double.parseDouble(s.split("\\s")[0]);
-    }
-    void setBBX(double x) {//下の吹き出しのX座標
-        String s = currentAction.getBottomBalloonOffset();
-        currentAction.setBottomBalloonOffset(""+x+" "+s.split("\\s")[1]);
-    }
-    double getBBY() {//下の吹き出しのY座標
-        String s = currentAction.getBottomBalloonOffset();
-        return Double.parseDouble(s.split("\\s")[1]);
-    }
-    void setBBY(double y) {//上の吹き出しのY座標
-        String s = currentAction.getBottomBalloonOffset();
-        currentAction.setBottomBalloonOffset(s.split("\\s")[0]+" "+y);
-    }
-    double getLX() {//ラベルのX座標
-        String s = currentAction.getLabelOffset();
-        return Double.parseDouble(s.split("\\s")[0]);
-    }
-    void setLX(double x) {//ラベルのX座標
-        String s = currentAction.getLabelOffset();
-        currentAction.setLabelOffset(""+x+" "+s.split("\\s")[1]);
-    }
-    double getLY() {//ラベルのY座標
-        String s = currentAction.getLabelOffset();
-        return Double.parseDouble(s.split("\\s")[1]);
-    }
-    void setLY(double y) {//ラベルのY座標
-        String s = currentAction.getLabelOffset();
-        currentAction.setLabelOffset(s.split("\\s")[0]+" "+y);
-    }
-    String[] getAllBones() {
-        if (currentSkeleton==null)
-            return new String[0];
-        else
-            return currentSkeleton.getBones();
+    void setLabelOffset(double v[]) {//ラベル
+        currentAction.setLabelOffset(vector2d2string(v));
     }
     String[] getUsedBones() {
         ArrayList<String> al = new ArrayList<String>();
@@ -396,9 +268,6 @@ class A3eActionEditor {
             return fileManager.getA3eFileFromPath(wrl);
         return A3eFile.noneFile;
     }
-    CenteredVRML getVRML(String bn) {
-        return vrmls.get(bn);
-    }
     void setVRML(String bn,A3eFile file) {
         P p = null;
         for (P pp : currentAction.getP()) {
@@ -421,7 +290,6 @@ class A3eActionEditor {
                 currentAction.getP().add(p);
             }
         }
-        setCurrent(currentAction);
     }
     void removeSoundDate() {
         currentAction.setS(null);
@@ -486,11 +354,7 @@ class A3eActionEditor {
         currentAction.getS().setContinue(b);
     }
     Vector3d getSoundOffset() {
-        String so = currentAction.getS().getOffset();
-        double x = Double.parseDouble(so.split("\\s")[0]);
-        double y = Double.parseDouble(so.split("\\s")[1]);
-        double z = Double.parseDouble(so.split("\\s")[2]);
-        return new Vector3d(x,y,z);
+        return string2vector3d(currentAction.getS().getOffset());
     }
     void setSoundOffset(double x,double y,double z) {
         S s = currentAction.getS();
@@ -500,11 +364,7 @@ class A3eActionEditor {
         s.setOffset(so);
     }
     Vector3d getSoundDirection() {
-        String sd = currentAction.getS().getDirection();
-        double x = Double.parseDouble(sd.split("\\s")[0]);
-        double y = Double.parseDouble(sd.split("\\s")[1]);
-        double z = Double.parseDouble(sd.split("\\s")[2]);
-        return new Vector3d(x,y,z);
+        return string2vector3d(currentAction.getS().getDirection());
     }
     void setSoundDirection(double x,double y,double z) {
         S s = currentAction.getS();
@@ -514,26 +374,28 @@ class A3eActionEditor {
         s.setDirection(sd);
     }
     Vector3d getOffset() {
-        String o = currentAction.getOffset();
-        double x = Double.parseDouble(o.split("\\s")[0]);
-        double y = Double.parseDouble(o.split("\\s")[1]);
-        double z = Double.parseDouble(o.split("\\s")[2]);
-        return new Vector3d(x,y,z);
+        return string2vector3d(currentAction.getOffset());
     }
     void setOffset(Vector3d v) {
-        String o = ""+v.x+" "+v.y+" "+v.z;
-        currentAction.setOffset(o);
+        currentAction.setOffset(vector3d2string(v));
     }
+    //返ってくるのはデグリー
     Vector3d getRot() {
-        String r = currentAction.getRot();
-        double x = Double.parseDouble(r.split("\\s")[0])/180.0*Math.PI;
-        double y = Double.parseDouble(r.split("\\s")[1])/180.0*Math.PI;
-        double z = Double.parseDouble(r.split("\\s")[2])/180.0*Math.PI;
-        return new Vector3d(x,y,z);
+        return string2vector3d(currentAction.getRot());
     }
+    //引数はデグリーで与えること
     void setRot(Vector3d v) {
-        String r = ""+180.0*v.x/Math.PI+" "+180.0*v.y/Math.PI+" "+180.0*v.z/Math.PI;
-        currentAction.setRot(r);
+        currentAction.setRot(vector3d2string(v));
+    }
+    Quat4d getQuat() {
+        Vector3d v = string2vector3d(currentAction.getRot());
+        v.scale(Math.PI/180.0);
+        return Util.euler2quat(v);
+    }
+    void setQuat(Quat4d quat) {
+        Vector3d v = Util.quat2euler(quat);
+        v.scale(180.0/Math.PI);
+        currentAction.setRot(vector3d2string(v));
     }
     double getScale() {
         return currentAction.getScale();
@@ -541,6 +403,41 @@ class A3eActionEditor {
     void setScale(double s) {
         currentAction.setScale(s);
     }
+
+
+    //返り値はCATALOGの生データ
+    T getPartData(String bn) {
+        P part = null;
+        for (P p : currentAction.getP()) {
+            if (p.getName().equals(bn)) {
+                part = p;
+                break;
+            }
+        }
+        if (part==null)
+            return null;
+        T t = new T();
+        t.offset = string2vector3d(part.getOffset());
+        t.rot = string2vector3d(part.getRot());
+        t.scale = part.getScale();
+        return t;
+    }
+    //引数はCATALOG.XMLにそのまま書き込まれる
+    void setPartData(String bn,T t) {
+        P part = null;
+        for (P p : currentAction.getP()) {
+            if (p.getName().equals(bn)) {
+                part = p;
+                break;
+            }
+        }
+        if (part==null)
+            return;
+        part.setOffset(vector3d2string(t.offset));
+        part.setRot(vector3d2string(t.rot));
+        part.setScale(t.scale);
+    }
+    /*
     void updateActionOffsetRotScale() {
         if (currentSkeleton==null)
             return;
@@ -565,6 +462,8 @@ class A3eActionEditor {
             actionQuatRev.conjugate();
             actionScale = currentSkeleton.getScale();
             bvhOffset = currentSkeleton.getOffset(bn);
+            if (bvhOffset==null) // bnが[none]の場合(だけだと思う)
+                bvhOffset = new Vector3d();
         } else {
             actionLoc = new Vector3d();
             actionQuatRev = new Quat4d(0.0,0.0,0.0,1.0);
@@ -579,18 +478,18 @@ class A3eActionEditor {
                 v.scale(1.0/actionScale);
                 v = Util.trans(actionQuatRev,v);
                 v.sub(bvhOffset);
-                String s = ""+v.x+" "+v.y+" "+v.z;
-                p.setOffset(s);
+                p.setOffset(vector3d2string(v));
 
                 Quat4d boneQuat = vrml.getQuat();
                 boneQuat.mul(actionQuatRev,boneQuat);
                 v = Util.quat2euler(boneQuat);
-                s = ""+180.0*v.x/Math.PI+" "+180.0*v.y/Math.PI+" "+180.0*v.z/Math.PI;
-                p.setRot(s);
+                v.scale(180.0/Math.PI);
+                p.setRot(vector3d2string(v));
                 p.setScale(vrml.getScale()/actionScale);
             }
         }
     }
+    */
     void saveCatalog() {
         try {
             editor.catalog.save();
@@ -599,4 +498,28 @@ class A3eActionEditor {
             e.printStackTrace();
         }
     }
+    Vector3d string2vector3d(String s) {
+        String ss[] = s.split("\\s+");
+        double x = Double.parseDouble(ss[0]);
+        double y = Double.parseDouble(ss[1]);
+        double z = Double.parseDouble(ss[2]);
+        return new Vector3d(x,y,z);
+    }
+    String vector3d2string(Vector3d v) {
+        return ""+v.x+" "+v.y+" "+v.z;
+    }
+    double[] string2vector2d(String s) {
+        String ss[] = s.split("\\s+");
+        double ret[] = new double[2];
+        ret[0] = Double.parseDouble(ss[0]);
+        ret[1] = Double.parseDouble(ss[1]);
+        return ret;
+    }
+    String vector2d2string(double v[]) {
+        return ""+v[0]+" "+v[1];
+    }
+}
+//Offset,Rot,Scaleをひとまとめにしたクラス
+class T {
+    Vector3d offset; Vector3d rot; double scale;
 }
