@@ -10,6 +10,7 @@ import javax.vecmath.Vector3d;
 
 import java.awt.*;
 import java.awt.event.*;
+import java.util.*;
 
 import jp.sourceforge.acerola3d.a3.*;
 import jp.sourceforge.acerola3d.a3panels.*;
@@ -21,6 +22,10 @@ class A3eActionEditorGUI extends JPanel implements ActionListener {
     JToggleButton viewMode; //めんどうなのでJToggleButtonにしちゃった
     double canvasWidthInPWorld = 1.0;
     MyExamController myExamController;
+    Skeleton skeleton;
+    HashMap<String,CenteredVRML> vrmls = new HashMap<String,CenteredVRML>();
+    ArrayList<Edit> editSequence = new ArrayList<Edit>();
+    int editSequenceIndex = -1;
 
     ////GUI
     //leftBox
@@ -60,6 +65,8 @@ class A3eActionEditorGUI extends JPanel implements ActionListener {
     JComboBox boneCB;
     JLabel vrmlL;
     JComboBox vrmlCB;
+    JButton undoB;
+    JButton redoB;
 
     Action3D measure;
     Action3D empty;
@@ -188,7 +195,7 @@ class A3eActionEditorGUI extends JPanel implements ActionListener {
         topBox.myAdd(topRBox,1);
         canvas = A3Canvas.createA3Canvas(300,300);
         MyMouseListener ml = new MyMouseListener(this);
-        canvas.addMouseListener(ml);
+        canvas.addA3Listener(ml);
         canvas.addMouseMotionListener(ml);
         centerBox.myAdd(canvas,1);
         //topLBox
@@ -286,6 +293,15 @@ class A3eActionEditorGUI extends JPanel implements ActionListener {
 
         rightBox.myAdd(Box.createVerticalGlue(),1);
 
+        MyHBox undoRedoBox = new MyHBox();
+        rightBox.myAdd(undoRedoBox,0);
+        undoB = new JButton("undo");
+        undoB.addActionListener(this);
+        undoRedoBox.myAdd(undoB,1);
+        redoB = new JButton("redo");
+        redoB.addActionListener(this);
+        undoRedoBox.myAdd(redoB,1);
+
         try {
             Action3D.clearCash("x-res:///jp/sourceforge/acerola3d/a3panels/resources/measure.a3");
             measure = new Action3D("x-res:///jp/sourceforge/acerola3d/a3panels/resources/measure.a3");
@@ -305,6 +321,7 @@ class A3eActionEditorGUI extends JPanel implements ActionListener {
     void keyboardShotcutInit() {
         ActionMap am = getActionMap();
         InputMap im = getInputMap(JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
+
         AbstractAction topViewAction = new AbstractAction() {
             private static final long serialVersionUID = 1L;
             public void actionPerformed(ActionEvent event) {
@@ -316,6 +333,7 @@ class A3eActionEditorGUI extends JPanel implements ActionListener {
         };
         am.put("topViewAction",topViewAction);
         im.put(KeyStroke.getKeyStroke('7'),"topViewAction");
+
         AbstractAction camViewAction = new AbstractAction() {
             private static final long serialVersionUID = 1L;
             public void actionPerformed(ActionEvent event) {
@@ -327,6 +345,7 @@ class A3eActionEditorGUI extends JPanel implements ActionListener {
         };
         am.put("camViewAction",camViewAction);
         im.put(KeyStroke.getKeyStroke('9'),"camViewAction");
+
         AbstractAction frontViewAction = new AbstractAction() {
             private static final long serialVersionUID = 1L;
             public void actionPerformed(ActionEvent event) {
@@ -338,6 +357,7 @@ class A3eActionEditorGUI extends JPanel implements ActionListener {
         };
         am.put("frontViewAction",frontViewAction);
         im.put(KeyStroke.getKeyStroke('1'),"frontViewAction");
+
         AbstractAction sideViewAction = new AbstractAction() {
             private static final long serialVersionUID = 1L;
             public void actionPerformed(ActionEvent event) {
@@ -349,8 +369,76 @@ class A3eActionEditorGUI extends JPanel implements ActionListener {
         };
         am.put("sideViewAction",sideViewAction);
         im.put(KeyStroke.getKeyStroke('3'),"sideViewAction");
+
+        AbstractAction undoAction = new AbstractAction() {
+            private static final long serialVersionUID = 1L;
+            public void actionPerformed(ActionEvent event) {
+                if (editSequenceIndex<0)
+                    return;
+                Edit e = editSequence.get(editSequenceIndex);
+                e.undo();
+                editSequenceIndex--;
+            }
+        };
+        am.put("undoAction",undoAction);
+        im.put(KeyStroke.getKeyStroke("ctrl Z"),"undoAction");
+
+        AbstractAction redoAction = new AbstractAction() {
+            private static final long serialVersionUID = 1L;
+            public void actionPerformed(ActionEvent event) {
+                if (editSequenceIndex>=(editSequence.size()-1))
+                    return;
+                editSequenceIndex++;
+                Edit e = editSequence.get(editSequenceIndex);
+                e.redo();
+            }
+        };
+        am.put("redoAction",redoAction);
+        im.put(KeyStroke.getKeyStroke("ctrl Y"),"redoAction");
+    }
+    void doEdit(Edit e) {
+        e.edit();
+        while ((editSequence.size()-1)>editSequenceIndex) {
+            editSequence.remove(editSequence.size()-1);
+        }
+        editSequence.add(e);
+        editSequenceIndex++;
     }
     void init() {
+        editSequence.clear();
+        editSequenceIndex=-1;
+        //==================================
+        try {
+            String bvhPath = actionEditor.getBVHPath();
+            String bvhUrl = actionEditor.getBVHURL();
+            if (bvhUrl!=null && !bvhPath.equals("none"))
+                skeleton = new Skeleton(bvhUrl);
+            else
+                skeleton = null;
+            Vector3d actionOffset = actionEditor.getOffset();
+            Quat4d actionQuat = actionEditor.getQuat();
+            double actionScale = actionEditor.getScale();
+            if (skeleton!=null) {
+                skeleton.setScale(actionScale);
+                skeleton.setQuat(actionQuat);
+                skeleton.setLoc(actionOffset);
+            }
+            //VRML.clearCash();
+            vrmls.clear();
+            for (String bn : actionEditor.getUsedBones()) {
+                String vrmlURL = actionEditor.getVRMLFile(bn).getURLString();
+                CenteredVRML vrml = new CenteredVRML(vrmlURL);
+                T t = actionEditor.getPartData(bn);
+                t = catalog2CenteredVRML(bn,t);
+                vrml.setScale(t.scale);
+                vrml.setRot(t.rot);
+                vrml.setLoc2(t.offset);
+                vrmls.put(bn,vrml);
+            }
+        } catch (Exception e) {
+            e.printStackTrace();
+        }
+        //==================================
         actionCB.removeActionListener(this);
         actionCB.removeAllItems();
         for (String an : actionEditor.getActionNames()) {
@@ -417,8 +505,11 @@ class A3eActionEditorGUI extends JPanel implements ActionListener {
         measureCB.setSelected(false);
         boneCB.removeAllItems();
         boneCB.addItem("[none]");
-        for (String bn : actionEditor.getAllBones()) {
-            boneCB.addItem(bn);
+        String allBones[] = {};
+        if (skeleton!=null) {
+            for (String bn : skeleton.getBones()) {
+                boneCB.addItem(bn);
+            }
         }
         vrmlL.setText("[none]のvrml:");
         vrmlCB.removeActionListener(this);
@@ -436,41 +527,38 @@ class A3eActionEditorGUI extends JPanel implements ActionListener {
         frontB.setSelected(true);
         canvas.setProjectionMode(ProjectionMode.PARALLEL);
         canvas.setNavigationMode(A3CanvasInterface.NaviMode.EDIT);
-        Skeleton s = actionEditor.getCurrentSkeleton();
-        if (s!=null)
-            canvas.add(s);
-        for (CenteredVRML vrml : actionEditor.getCurrentVRMLs()) {
+        if (skeleton!=null)
+            canvas.add(skeleton);
+        for (CenteredVRML vrml : vrmls.values()) {
             canvas.add(vrml);
         }
-        double x,y;
-        x = actionEditor.getRBX(); y = actionEditor.getRBY();
-        empty.setBalloonLoc(A3Object.BalloonDir.RIGHT, x, y);
-        x = actionEditor.getLBX(); y = actionEditor.getLBY();
-        empty.setBalloonLoc(A3Object.BalloonDir.LEFT, x, y);
-        x = actionEditor.getTBX(); y = actionEditor.getTBY();
-        empty.setBalloonLoc(A3Object.BalloonDir.TOP, x, y);
-        x = actionEditor.getBBX(); y = actionEditor.getBBY();
-        empty.setBalloonLoc(A3Object.BalloonDir.BOTTOM, x, y);
-        x = actionEditor.getLX(); y = actionEditor.getLY();
-        empty.setLabelLoc(x, y);
+        double xy[];
+        xy = actionEditor.getRightBalloonOffset();
+        empty.setBalloonLoc(A3Object.BalloonDir.RIGHT, xy[0], xy[1]);
+        xy = actionEditor.getLeftBalloonOffset();
+        empty.setBalloonLoc(A3Object.BalloonDir.LEFT, xy[0], xy[1]);
+        xy = actionEditor.getTopBalloonOffset();
+        empty.setBalloonLoc(A3Object.BalloonDir.TOP, xy[0], xy[1]);
+        xy = actionEditor.getBottomBalloonOffset();
+        empty.setBalloonLoc(A3Object.BalloonDir.BOTTOM, xy[0], xy[1]);
+        xy = actionEditor.getLabelOffset();
+        empty.setLabelLoc(xy[0], xy[1]);
         adjustFront();
         this.requestFocus();
     }
     BoundingSphere getAllBound() {
         BoundingSphere ret = null;
-        CenteredVRML vrmls[] = actionEditor.getCurrentVRMLs();
-        for (CenteredVRML v : vrmls) {
+        for (CenteredVRML v : vrmls.values()) {
             if (ret==null)
                 ret = new BoundingSphere(v.getNode().getBounds());
             else
                 ret.combine(v.getNode().getBounds());
         }
-        Skeleton s = actionEditor.getCurrentSkeleton();
-        if (s!=null) {
+        if (skeleton!=null) {
             if (ret==null)
-                ret = new BoundingSphere(s.getNode().getBounds());
+                ret = new BoundingSphere(skeleton.getNode().getBounds());
             else
-                ret.combine(s.getNode().getBounds());
+                ret.combine(skeleton.getNode().getBounds());
         }
         if (ret==null)
             ret = new BoundingSphere();
@@ -768,20 +856,18 @@ class A3eActionEditorGUI extends JPanel implements ActionListener {
             String m = (String)editModeCB.getSelectedItem();
             if (m.equals("全部")) {
                 canvas.delAll();
-                Skeleton skeleton = actionEditor.getCurrentSkeleton();
                 if (skeleton!=null) {
                     canvas.add(skeleton);
                     skeleton.setPickable(true);
                     //skeleton.unpolygonize();
                 }
-                for (CenteredVRML vrml : actionEditor.getCurrentVRMLs()) {
+                for (CenteredVRML vrml : vrmls.values()) {
                     canvas.add(vrml);
                     vrml.setPickable(true);
                     //vrml.unpolygonize();
                 }
             } else if (m.equals("骨だけ")) {
                 canvas.delAll();
-                Skeleton skeleton = actionEditor.getCurrentSkeleton();
                 if (skeleton!=null) {
                     canvas.add(skeleton);
                     skeleton.setPickable(true);
@@ -789,27 +875,25 @@ class A3eActionEditorGUI extends JPanel implements ActionListener {
                 }
             } else if (m.equals("形すべて")) {
                 canvas.delAll();
-                Skeleton skeleton = actionEditor.getCurrentSkeleton();
                 if (skeleton!=null) {
                     canvas.add(skeleton);
                     skeleton.setPickable(false);
                     //skeleton.unpolygonize();
                 }
-                for (CenteredVRML vrml : actionEditor.getCurrentVRMLs()) {
+                for (CenteredVRML vrml : vrmls.values()) {
                     canvas.add(vrml);
                     vrml.setPickable(true);
                     //vrml.polygonize();
                 }
             } else if (m.equals("一つの形")) {
                 canvas.delAll();
-                Skeleton skeleton = actionEditor.getCurrentSkeleton(); 
                 if (skeleton!=null) {
                     canvas.add(skeleton);
                     skeleton.setPickable(false);
                     //skeleton.unpolygonize();
                 }
                 String bn = (String)boneCB.getSelectedItem();
-                CenteredVRML vrml = actionEditor.getVRML(bn);
+                CenteredVRML vrml = vrmls.get(bn);
                 if (vrml!=null) {
                     canvas.add(vrml);
                     vrml.setPickable(true);
@@ -817,13 +901,12 @@ class A3eActionEditorGUI extends JPanel implements ActionListener {
                 }
             } else if (m.equals("上吹き出し")) {
                 canvas.delAll();
-                Skeleton skeleton = actionEditor.getCurrentSkeleton();
                 if (skeleton!=null) {
                     canvas.add(skeleton);
                     skeleton.setPickable(false);
                     //skeleton.unpolygonize();
                 }
-                for (CenteredVRML vrml : actionEditor.getCurrentVRMLs()) {
+                for (CenteredVRML vrml : vrmls.values()) {
                     canvas.add(vrml);
                     vrml.setPickable(false);
                     //vrml.unpolygonize();
@@ -834,13 +917,12 @@ class A3eActionEditorGUI extends JPanel implements ActionListener {
                 empty.setLabel(null);
             } else if (m.equals("下吹き出し")) {
                 canvas.delAll();
-                Skeleton skeleton = actionEditor.getCurrentSkeleton();
                 if (skeleton!=null) {
                     canvas.add(skeleton);
                     skeleton.setPickable(false);
                     //skeleton.unpolygonize();
                 }
-                for (CenteredVRML vrml : actionEditor.getCurrentVRMLs()) {
+                for (CenteredVRML vrml : vrmls.values()) {
                     canvas.add(vrml);
                     vrml.setPickable(false);
                     //vrml.unpolygonize();
@@ -851,13 +933,12 @@ class A3eActionEditorGUI extends JPanel implements ActionListener {
                 empty.setLabel(null);
             } else if (m.equals("右吹き出し")) {
                 canvas.delAll();
-                Skeleton skeleton = actionEditor.getCurrentSkeleton();
                 if (skeleton!=null) {
                     canvas.add(skeleton);
                     skeleton.setPickable(false);
                     //skeleton.unpolygonize();
                 }
-                for (CenteredVRML vrml : actionEditor.getCurrentVRMLs()) {
+                for (CenteredVRML vrml : vrmls.values()) {
                     canvas.add(vrml);
                     vrml.setPickable(false);
                     //vrml.unpolygonize();
@@ -868,13 +949,12 @@ class A3eActionEditorGUI extends JPanel implements ActionListener {
                 empty.setLabel(null);
             } else if (m.equals("左吹き出し")) {
                 canvas.delAll();
-                Skeleton skeleton = actionEditor.getCurrentSkeleton();
                 if (skeleton!=null) {
                     canvas.add(skeleton);
                     skeleton.setPickable(false);
                     //skeleton.unpolygonize();
                 }
-                for (CenteredVRML vrml : actionEditor.getCurrentVRMLs()) {
+                for (CenteredVRML vrml : vrmls.values()) {
                     canvas.add(vrml);
                     vrml.setPickable(false);
                     //vrml.unpolygonize();
@@ -885,13 +965,12 @@ class A3eActionEditorGUI extends JPanel implements ActionListener {
                 empty.setLabel(null);
             } else if (m.equals("ラベル")) {
                 canvas.delAll();
-                Skeleton skeleton = actionEditor.getCurrentSkeleton();
                 if (skeleton!=null) {
                     canvas.add(skeleton);
                     skeleton.setPickable(false);
                     //skeleton.unpolygonize();
                 }
-                for (CenteredVRML vrml : actionEditor.getCurrentVRMLs()) {
+                for (CenteredVRML vrml : vrmls.values()) {
                     canvas.add(vrml);
                     vrml.setPickable(false);
                     //vrml.unpolygonize();
@@ -901,7 +980,6 @@ class A3eActionEditorGUI extends JPanel implements ActionListener {
                 empty.setLabel("ラベル");
             } else if (m.equals("サウンド")) {
                 canvas.delAll();
-                Skeleton skeleton = actionEditor.getCurrentSkeleton();
                 if (skeleton!=null) {
                     canvas.add(skeleton);
                     skeleton.setPickable(false);
@@ -929,19 +1007,18 @@ class A3eActionEditorGUI extends JPanel implements ActionListener {
             vrmlCB.removeActionListener(this);
             vrmlCB.setSelectedItem(actionEditor.getVRMLFile(bn));
             vrmlCB.addActionListener(this);
-            Skeleton st = actionEditor.getCurrentSkeleton();
-            if (st!=null) {
-                st.resetSelected();
-                st.setBoneSelected(bn,true);
+            if (skeleton!=null) {
+                skeleton.resetSelected();
+                skeleton.setBoneSelected(bn,true);
             }
             if (editModeCB.getSelectedItem().equals("一つの形")) {
                 canvas.delAll();
-                if (st!=null) {
-                    canvas.add(st);
-                    st.setPickable(false);
-                    //st.unpolygonize();
+                if (skeleton!=null) {
+                    canvas.add(skeleton);
+                    skeleton.setPickable(false);
+                    //skeleton.unpolygonize();
                 }
-                CenteredVRML vrml = actionEditor.getVRML(bn);
+                CenteredVRML vrml = vrmls.get(bn);
                 if (vrml!=null) {
                     canvas.add(vrml);
                     vrml.setPickable(true);
@@ -949,26 +1026,121 @@ class A3eActionEditorGUI extends JPanel implements ActionListener {
                 }
             }
         } else if (s==vrmlCB) {
-            String bn = (String)boneCB.getSelectedItem();
-            A3eFile wrl = (A3eFile)vrmlCB.getSelectedItem();
-            actionEditor.setVRML(bn,wrl);
-            init();
+            doEdit(new VrmlChangeEdit());
+        } else if (s==undoB) {
+            if (editSequenceIndex>=0) {
+                Edit e = editSequence.get(editSequenceIndex);
+                e.undo();
+                editSequenceIndex--;
+            }
+        } else if (s==redoB) {
+            if (editSequenceIndex<(editSequence.size()-1)) {
+                editSequenceIndex++;
+                Edit e = editSequence.get(editSequenceIndex);
+                e.redo();
+            }
         } else {
             System.out.println("gaha?????");
         }
     }
-    class MyMouseListener extends MouseAdapter implements MouseMotionListener {
+
+
+
+
+    //CenteredVRMLの値からCATALOG.XMLに保存するための値を計算する(Offset,Rot,Scaleの全部)。
+    //この計算にはカレントアクションの変換が反映される。それとSkeletonがある場合には、そのSkeleton
+    //のBoneの変換も考慮する。Skeletonが無い場合などは最初の引数のbn(BoneName)は無視される。
+    //それとOffsetの値はCenteredVRMLのgetLoc2,setLoc2で使う値とする。
+    T centeredVRML2CATALOG(String bn,T t) {
+        Vector3d actionOffset = actionEditor.getOffset();
+        Quat4d actionQuatRev = actionEditor.getQuat();
+        actionQuatRev.conjugate();
+        double actionScale = actionEditor.getScale();
+        Vector3d bvhOffset;
+        if (skeleton!=null) {
+            bvhOffset = skeleton.getOffset(bn);
+            if (bvhOffset==null) // bnが[none]の場合(だけだと思う)
+                bvhOffset = new Vector3d();
+        } else {
+            bvhOffset = new Vector3d();
+        }
+        T tt = new T();
+        tt.offset = new Vector3d(t.offset);
+        tt.offset.sub(actionOffset);
+        tt.offset.scale(1.0/actionScale);
+        tt.offset = Util.trans(actionQuatRev,tt.offset);
+        tt.offset.sub(bvhOffset);
+
+        Quat4d boneQuat = Util.euler2quat(t.rot);
+        boneQuat.mul(actionQuatRev,boneQuat);
+        tt.rot = Util.quat2euler(boneQuat);
+        tt.rot.scale(180.0/Math.PI);
+        tt.scale = t.scale/actionScale;
+        return tt;
+    }
+
+    //CATALOG.XMLに保存された値からCenteredVRMLにセットするための値を計算する(Offset,Rot,Scaleの全部)。
+    //この計算にはカレントアクションの変換も考慮されている。Skeletonがある場合には
+    //Boneの変換も考慮される。Skeletonが無い場合には最初の引数bn(BoneName)は無視される。
+    //それと返り値のOffsetの値はCenteredVRMLのgetLoc2,setLoc2で使う値とする。
+    T catalog2CenteredVRML(String bn,T t) {
+        Vector3d actionOffset = actionEditor.getOffset();
+        Quat4d actionQuat = actionEditor.getQuat();
+        double actionScale = actionEditor.getScale();
+        Vector3d bvhOffset;
+        if (skeleton!=null) {
+            bvhOffset = skeleton.getOffset(bn);
+            if (bvhOffset==null) // bnが[none]の場合(だけだと思う)
+                bvhOffset = new Vector3d();
+        } else {
+            bvhOffset = new Vector3d();
+        }
+        Vector3d boneOffset = new Vector3d(t.offset);//CATALOG.XMLから来る骨のオフセット
+        Vector3d v = new Vector3d(t.rot);
+        v.scale(Math.PI/180.0);
+        Quat4d boneQuat = Util.euler2quat(v);//CATALOC.XMLから来る骨の回転
+        double boneScale = t.scale;//CATALOG.XMLから来る骨のスケール
+
+        T tt = new T();
+        tt.offset = new Vector3d(boneOffset);
+        tt.offset.add(bvhOffset);
+        tt.offset.scale(actionScale);
+        tt.offset = Util.trans(actionQuat,tt.offset);
+        tt.offset.add(actionOffset);
+        Quat4d quat = new Quat4d();
+        quat.mul(actionQuat,boneQuat);
+        tt.rot = Util.quat2euler(quat);
+        tt.scale = boneScale*actionScale;
+        return tt;
+    }
+
+
+    String getBoneName(CenteredVRML vrml) {
+        String boneName = null;
+        for (String bn : vrmls.keySet()) {
+            CenteredVRML v = vrmls.get(bn);
+            if (v==vrml) {
+                boneName = bn;
+                break;
+            }
+        }
+        return boneName;
+    }
+
+    class MyMouseListener extends A3Adapter implements MouseMotionListener {
         A3eActionEditorGUI actionEditorGUI;
         double preX;
         double preY;
+        A3Object targetA3;
         MyMouseListener(A3eActionEditorGUI aeGUI) {
             actionEditorGUI = aeGUI;
         }
         @Override
-        public void mousePressed(MouseEvent e) {
-            int x = e.getX(); int y = e.getY();
+        public void mousePressed(A3Event e) {
+            int x = e.getMouseEvent().getX(); int y = e.getMouseEvent().getY();
             Point3d p = canvas.canvasToPhysicalCS(x,y,1.0);
             preX = p.x; preY = p.y;
+            targetA3 = e.getA3Object();
         }
         @Override
         public void mouseDragged(MouseEvent e) {
@@ -1011,44 +1183,41 @@ class A3eActionEditorGUI extends JPanel implements ActionListener {
             preX = nowX; preY = nowY;
         }
         @Override
-        public void mouseReleased(MouseEvent e) {
-            mouseDragged(e);
+        public void mouseMoved(MouseEvent e) {;}
+        @Override
+        public void mouseReleased(A3Event e) {
+            mouseDragged(e.getMouseEvent());
             String m = (String)editModeCB.getSelectedItem();
+            double xy[] = new double[2];
             if (m.equals("全部")) {
-                actionEditor.updateActionOffsetRotScale();
-                actionEditor.updatePartsOffsetRotScale();
+                if (targetA3 instanceof Skeleton)
+                    updateActionOffsetRotScale();
+                else if (targetA3 instanceof CenteredVRML)
+                    updatePartsOffsetRotScale((CenteredVRML)targetA3);
             } else if (m.equals("骨だけ")) {
-                actionEditor.updateActionOffsetRotScale();
+                if (targetA3 instanceof Skeleton)
+                    updateActionOffsetRotScale();
             } else if (m.equals("形すべて")) {
-                actionEditor.updatePartsOffsetRotScale();
+                if (targetA3 instanceof CenteredVRML)
+                    updatePartsOffsetRotScale((CenteredVRML)targetA3);
             } else if (m.equals("一つの形")) {
-                String bn = (String)boneCB.getSelectedItem();
-                actionEditor.updatePartOffsetRotScale(bn);
+                if (targetA3 instanceof CenteredVRML)
+                    updatePartsOffsetRotScale((CenteredVRML)targetA3);
             } else if (m.equals("上吹き出し")) {
-                double xy[] = new double[2];
                 empty.getBalloonLoc(A3Object.BalloonDir.TOP,xy);
-                actionEditor.setTBX(xy[0]);
-                actionEditor.setTBY(xy[1]);
+                actionEditor.setTopBalloonOffset(xy);
             } else if (m.equals("下吹き出し")) {
-                double xy[] = new double[2];
                 empty.getBalloonLoc(A3Object.BalloonDir.BOTTOM,xy);
-                actionEditor.setBBX(xy[0]);
-                actionEditor.setBBY(xy[1]);
+                actionEditor.setBottomBalloonOffset(xy);
             } else if (m.equals("右吹き出し")) {
-                double xy[] = new double[2];
                 empty.getBalloonLoc(A3Object.BalloonDir.RIGHT,xy);
-                actionEditor.setRBX(xy[0]);
-                actionEditor.setRBY(xy[1]);
+                actionEditor.setRightBalloonOffset(xy);
             } else if (m.equals("左吹き出し")) {
-                double xy[] = new double[2];
                 empty.getBalloonLoc(A3Object.BalloonDir.LEFT,xy);
-                actionEditor.setLBX(xy[0]);
-                actionEditor.setLBY(xy[1]);
+                actionEditor.setLeftBalloonOffset(xy);
             } else if (m.equals("ラベル")) {
-                double xy[] = new double[2];
                 empty.getLabelLoc(xy);
-                actionEditor.setLX(xy[0]);
-                actionEditor.setLY(xy[1]);
+                actionEditor.setLabelOffset(xy);
             } else if (m.equals("サウンド")) {
                 Vector3d v = speaker.getLoc();
                 actionEditor.setSoundOffset(v.x,v.y,v.z);
@@ -1059,6 +1228,14 @@ class A3eActionEditorGUI extends JPanel implements ActionListener {
                 System.out.println("???????"+m);
             }
         }
+        void updateActionOffsetRotScale() {
+            if (skeleton!=null) {
+                doEdit(new SkeletonEdit());
+            }
+        }
+        void updatePartsOffsetRotScale(CenteredVRML vrml) {
+            doEdit(new PartEdit(vrml));
+        }
     }
     void clear() {
         //TODO
@@ -1109,4 +1286,177 @@ class A3eActionEditorGUI extends JPanel implements ActionListener {
             bvhCB.setSelectedItem(A3eFile.noneFile);
         }
     }
+    //Undo機能とかを実現するためのクラス
+    abstract class Edit {
+        abstract void edit();
+        abstract void undo();
+        void redo() {
+            edit();
+        }
+    }
+
+    class SkeletonEdit extends Edit {
+        Vector3d oldOffset;
+        Vector3d oldRot;//ラジアンで保存
+        double oldScale;
+        Vector3d newOffset;
+        Vector3d newRot;//ラジアンで保存
+        double newScale;
+        SkeletonEdit() {
+            oldOffset = actionEditor.getOffset();
+            oldRot = actionEditor.getRot();
+            oldRot.scale(Math.PI/180.0);
+            oldScale = actionEditor.getScale();
+            newOffset = skeleton.getLoc();
+            newRot = skeleton.getRot();
+            newScale = skeleton.getScale();
+        }
+        void edit(){
+            actionEditor.setScale(newScale);
+            Vector3d v = new Vector3d(newRot);
+            v.scale(180.0/Math.PI);
+            actionEditor.setOffset(newOffset);
+            actionEditor.setRot(v);
+            skeleton.setScale(newScale);
+            skeleton.setRot(newRot);
+            skeleton.setLoc(newOffset);
+        }
+        void undo(){
+            actionEditor.setScale(oldScale);
+            Vector3d v = new Vector3d(oldRot);
+            v.scale(180.0/Math.PI);
+            actionEditor.setRot(v);
+            actionEditor.setOffset(oldOffset);
+            skeleton.setScale(oldScale);
+            skeleton.setRot(oldRot);
+            skeleton.setLoc(oldOffset);
+        }
+    }
+    //これに記録される値はCATALOG.XMLに記録される値ではなくCenteredVRML
+    //に直接渡せる値。かつOffsetはsetLocとかgetLocでなくsetLoc2とgetLoc2
+    //を使うことを前提にしている。
+    class PartEdit extends Edit {
+        String boneName;
+        Vector3d oldOffset = new Vector3d();
+        Vector3d oldRot = new Vector3d();
+        double oldScale;
+        Vector3d newOffset;
+        Vector3d newRot;
+        double newScale;
+        PartEdit(CenteredVRML target) {
+            boneName = getBoneName(target);
+            T t = actionEditor.getPartData(boneName);
+            t = catalog2CenteredVRML(boneName,t);
+            oldOffset = t.offset;
+            oldRot = t.rot;
+            oldScale = t.scale;
+            newOffset = target.getLoc2();
+            newRot = target.getRot();
+            newScale = target.getScale();
+        }
+        void edit(){
+            T t = new T();
+            t.offset = newOffset;
+            t.rot = newRot;
+            t.scale = newScale;
+            t = centeredVRML2CATALOG(boneName,t);
+            actionEditor.setPartData(boneName,t);
+            CenteredVRML target = vrmls.get(boneName);
+            target.setRot(newRot);
+            target.setScale(newScale);
+            target.setLoc2(newOffset);
+        }
+        void undo(){
+            T t = new T();
+            t.offset = oldOffset;
+            t.rot = oldRot;
+            t.scale = oldScale;
+            t = centeredVRML2CATALOG(boneName,t);
+            actionEditor.setPartData(boneName,t);
+            CenteredVRML target = vrmls.get(boneName);
+            target.setScale(oldScale);
+            target.setRot(oldRot);
+            target.setLoc2(oldOffset);
+        }
+    }
+    
+    class VrmlChangeEdit extends Edit {
+        String boneName;
+        A3eFile oldFile;
+        Vector3d offset;
+        Vector3d rot;
+        double scale;
+        A3eFile newFile;
+        VrmlChangeEdit() {
+            boneName = (String)boneCB.getSelectedItem();
+            oldFile = actionEditor.getVRMLFile(boneName);
+            newFile = (A3eFile)vrmlCB.getSelectedItem();
+            CenteredVRML oldVRML = vrmls.get(boneName);
+            if (oldVRML!=null) {
+                offset = oldVRML.getLoc2();
+                rot = oldVRML.getRot();
+                scale = oldVRML.getScale();
+            }
+        }
+        void edit(){
+            actionEditor.setVRML(boneName,newFile);
+            CenteredVRML oldVRML = vrmls.get(boneName);
+            CenteredVRML newVRML = null;
+            if (newFile==A3eFile.noneFile) {
+                newVRML = null;
+                vrmls.remove(boneName);
+            } else {
+                try {
+                    newVRML = new CenteredVRML(newFile.getURLString());
+                    vrmls.put(boneName,newVRML);
+                } catch (Exception ee) {
+                    ee.printStackTrace();
+                }
+            }
+            if (oldVRML!=null)
+                canvas.del(oldVRML);
+            if (newVRML!=null) {
+                newVRML.setScale(scale);
+                newVRML.setRot(rot);
+                newVRML.setLoc2(offset);
+                canvas.add(newVRML);
+            }
+            boneCB.removeActionListener(A3eActionEditorGUI.this);
+            vrmlCB.removeActionListener(A3eActionEditorGUI.this);
+            boneCB.setSelectedItem(boneName);
+            vrmlCB.setSelectedItem(newFile);
+            boneCB.addActionListener(A3eActionEditorGUI.this);
+            vrmlCB.addActionListener(A3eActionEditorGUI.this);
+        }
+        void undo(){
+            actionEditor.setVRML(boneName,oldFile);
+            CenteredVRML newVRML = vrmls.get(boneName);
+            CenteredVRML oldVRML = null;
+            if (oldFile==A3eFile.noneFile) {
+                oldVRML = null;
+                vrmls.remove(boneName);
+            } else {
+                try {
+                    oldVRML = new CenteredVRML(oldFile.getURLString());
+                    vrmls.put(boneName,oldVRML);
+                } catch (Exception ee) {
+                    ee.printStackTrace();
+                }
+            }
+            if (newVRML!=null)
+                canvas.del(newVRML);
+            if (oldVRML!=null) {
+                oldVRML.setScale(scale);
+                oldVRML.setRot(rot);
+                oldVRML.setLoc2(offset);
+                canvas.add(oldVRML);
+            }
+            boneCB.removeActionListener(A3eActionEditorGUI.this);
+            vrmlCB.removeActionListener(A3eActionEditorGUI.this);
+            boneCB.setSelectedItem(boneName);
+            vrmlCB.setSelectedItem(oldFile);
+            boneCB.addActionListener(A3eActionEditorGUI.this);
+            vrmlCB.addActionListener(A3eActionEditorGUI.this);
+        }
+    }
 }
